package edu.gmu.trimdroid.soot.callgraph;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;

import soot.*;
import soot.options.Options;
import edu.gmu.trimdroid.soot.callgraph.CGEdge.EdgeType;
import edu.gmu.trimdroid.soot.utils.Const;
import edu.gmu.trimdroid.soot.utils.Util;

public class BehaviorModelExtractor {

	Queue<CGNode> toBevisitedNodes;
	Set<CGNode> visited;
	CGNode root;
	SootClass rClazz;

    public BehaviorModelExtractor(String apkFile, String apkFolder, String outputPath, String fileName) throws JAXBException, IOException, SAXException, ParserConfigurationException {

        applySootOptions(apkFile);
        Scene.v().loadNecessaryClasses();
        PackManager.v().runPacks();
        toBevisitedNodes = new LinkedList<>();
        visited = new HashSet<>();
        rClazz = Util.findRClass("id");
        addEntryPoint(apkFolder);
        visitNodes();

        CGNode root = this.getRoot();
        CGNode.ignoreSimpleLinks = true;

        File folder = new File(outputPath);
        if(!folder.exists()) {
            folder.mkdir();
        }
        String output = outputPath+fileName;
        root.printCallGraph(output + ".gv");
        File file = new File(output+".xml");
        if(!file.exists())
            Util.toXML(output + ".xml", root);
    }

	public CGNode getRoot() {
		return root;
	}

	private void visitNodes() {
		while (!toBevisitedNodes.isEmpty()) {
			CGNode node = toBevisitedNodes.remove();
			visited.add(node);
			processNode(node);
		}
	}

	private void processNode(CGNode parent) {
		SootMethod method = parent.method;
		if (method == null || !method.getDeclaringClass().isApplicationClass())
			return;
		try {
			if (Arrays.asList(Const.ENTRY_POINT_MENU).contains(method.getName())) {
				addNodeToQueue(GUINodeExtractor.extractMenuNodes(method, parent, rClazz));
			} else {
				PatchingChain<Unit> units = method.getActiveBody().getUnits();
				for (Unit unit : units) {
					NodeExtractor[] nodeExtractors = { new GUINodeExtractor(), new IPCNodeExtractor(), new SimpleNodeExtractor() };
					for (NodeExtractor nodeExtractor : nodeExtractors) {
						addNodeToQueue(nodeExtractor.extractNode(unit, units, method, parent, this.rClazz));
					}
				}
			}
		} catch (RuntimeException e) {
			System.err.println(e.getMessage());
		}
	}

	private void addNodeToQueue(List<CGNode> nodes) {
		if (nodes == null)
			return;
		for (CGNode node : nodes) {
			addNodeToQueue(node);
		}
	}

	private void addNodeToQueue(CGNode node) {
		if (node != null && !visited.contains(node)) {
			toBevisitedNodes.add(node);
		}
	}

	private static void applySootOptions(String apk) {
		String androidHome = System.getenv("ANDROID_HOME");
        G.reset();
        Options.v().set_src_prec(Options.src_prec_apk);
		Options.v().set_output_format(Options.output_format_none);
		Options.v().set_no_bodies_for_excluded(true);
		Options.v().set_show_exception_dests(false);
		Options.v().set_allow_phantom_refs(true);
		Options.v().set_print_tags_in_output(false);
		Options.v().set_android_jars(androidHome+"/platforms");
		Options.v().set_process_dir(Collections.singletonList(apk));
	}

	private void addEntryPoint(String manifestFile) throws ParserConfigurationException, SAXException, IOException {
		root = new CGNode("Root");
		SootClass activityClass = Util.getMainActivityClass(manifestFile);
		for (String entryPointMethod : Const.ENTRY_POINT_METHODS) {
			try {
				SootMethod mainMethod = activityClass.getMethodByName(entryPointMethod);
				CGNode entryPoint = new CGNode(mainMethod);
				CGNode.connect(root, entryPoint, new CGEdge(EdgeType.SYMBOLIC), false);
				toBevisitedNodes.add(entryPoint);
			} catch (RuntimeException e) {
				System.err.println(e.getMessage());
			}
		}
	}
}
