package edu.gmu.trimdroid.soot.depextract;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import soot.Body;
import soot.Scene;
import soot.SootClass;
import soot.SootField;
import soot.SootMethod;
import soot.SootMethodRef;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.InvokeExpr;
import soot.jimple.StaticInvokeExpr;
import soot.jimple.internal.JInstanceFieldRef;
import soot.jimple.internal.JimpleLocal;
import edu.gmu.trimdroid.soot.utils.EventHandlerExtractor;
import edu.gmu.trimdroid.soot.utils.Util;

public class HandlerDependencyExtractor extends AbstractDependencyExtractor {

	Map<String, List<SootField>> handlerWidgetAccessMap;
	Map<String, Set<String>> handlerWidgetDirectAccessMap;
	Map<SootField, String> widgetIdFieldMap;

	public HandlerDependencyExtractor() {
		super();
		this.handlerWidgetDirectAccessMap = new HashMap<>();
		this.handlerWidgetAccessMap = new HashMap<>();
		this.widgetIdFieldMap = new HashMap<>();
	}

	@Override
	protected List<List<String>> getDependencies() {
		List<List<String>> result = new ArrayList<>();
		for (String handler : handlerWidgetAccessMap.keySet()) {
			Set<String> tmp = new HashSet<>();
			tmp.add(handler.split("\\[")[1].replace("]", "").trim());
			for (SootField field : handlerWidgetAccessMap.get(handler)) {
				String widgetId = widgetIdFieldMap.get(field);
				if (widgetId != null)
					tmp.add(widgetId.split("\\[")[1].replace("]", "").trim());
			}
			if (handlerWidgetDirectAccessMap.get(handler) != null && !handlerWidgetDirectAccessMap.get(handler).isEmpty())
				tmp.addAll(handlerWidgetDirectAccessMap.get(handler));
			if (tmp.size() > 1)
				result.add(new ArrayList<String>(tmp));
		}
		return result;
	}

	@Override
	public void doDependencyAnalysis(Body b) {
		findWidgetHandlers();
		findAllWidgetIds();
	}

	private void findAllWidgetIds() {
		for (Unit unit : units) {
			for (ValueBox useBox : unit.getUseBoxes()) {
				if (useBox.getValue() instanceof InvokeExpr) {
					InvokeExpr invokeExpr = (InvokeExpr) useBox.getValue();
					if (invokeExpr.getMethodRef().name().matches("^findViewById$")) {
						Value widgetId = invokeExpr.getArg(0);
						SootField field = findWidgetId(unit);
						if (field != null && widgetIdFieldMap.get(field) == null)
							widgetIdFieldMap.put(field, field.getType() + "[" + widgetId.toString() + "]");
					}
				}
			}
		}
	}

	private SootField findWidgetId(Unit usedPoint) {
		// Check if the widget assigned to local field
		for (ValueBox defBox : usedPoint.getDefBoxes()) {
			if (defBox.getValue() instanceof JInstanceFieldRef) {
				JInstanceFieldRef field = (JInstanceFieldRef) defBox.getValue();
				SootClass fieldClass = Scene.v().loadClass(field.getType().toString(), SootClass.SIGNATURES);
				if (Util.isInheritedFromClass(fieldClass, "android.view.View")) {
					return field.getField();
				}
			}
		}
		// Otherwise do forward search for finding widget
		for (ValueBox defBox : usedPoint.getDefBoxes()) {
			Unit usePoint = null;
			Value defVar = defBox.getValue();
			if (defVar instanceof JInstanceFieldRef) {
				JInstanceFieldRef ref = ((JInstanceFieldRef) defVar);
				usePoint = findClosestUsePointForward(usedPoint, ref.getFieldRef());
			}
			if (defVar instanceof JimpleLocal) {
				usePoint = findClosestUsePointForward(usedPoint, defVar);
			}
			if (usePoint != null) {
				SootField field = findWidgetId(usePoint);
				if (field != null)
					return field;
			}
		}
		return null;
	}

	private void findWidgetHandlers() {
		for (Unit unit : units) {
			Object[] widgetHandlerAndId = EventHandlerExtractor.findWidgetHandlerAndId(units, unit);
			SootMethod listener = (SootMethod) widgetHandlerAndId[0];
			Value widgetId = (Value) widgetHandlerAndId[1];
			SootMethodRef methodRef = (SootMethodRef) widgetHandlerAndId[2];
			if (listener != null && widgetId != null) {
				handlerWidgetAccessMap.put(methodRef.declaringClass().getName() + "[" + widgetId.toString() + "]", findUsedWidgetsAccessMethod(listener));
				findAllDirectAccesses(listener, methodRef.declaringClass().getName() + "[" + widgetId.toString() + "]", widgetId.toString());
			}
		}
	}

	private static List<SootField> findUsedWidgetsAccessMethod(SootMethod listener) {
		List<SootField> usedWidgets = new ArrayList<>();
		try {
			for (Unit unit : listener.getActiveBody().getUnits()) {
				for (ValueBox useBox : unit.getUseBoxes()) {
					if (useBox.getValue() instanceof StaticInvokeExpr) {
						SootMethod method = ((StaticInvokeExpr) useBox.getValue()).getMethod();
						if (method.getName().startsWith("access$")) {
							usedWidgets.add(findAccessMethods(method));
						}
					}
				}
			}
			System.out.println(listener);
		} catch (Exception e) {
			System.err.println(listener);
		}
		return usedWidgets;
	}

	private static SootField findAccessMethods(SootMethod accessMethod) {
		for (Unit unit : accessMethod.getActiveBody().getUnits()) {
			for (ValueBox useBox : unit.getUseBoxes()) {
				if (useBox.getValue() instanceof JInstanceFieldRef) {
					JInstanceFieldRef ref = ((JInstanceFieldRef) useBox.getValue());
					return ref.getField();
				}
			}
		}
		return null;
	}

	private void findAllDirectAccesses(SootMethod listener, String methodName, String handlerWidgetId) {
		try {
			Set<String> usedWidgets = new HashSet<>();
			for (Unit unit : listener.getActiveBody().getUnits()) {
				for (ValueBox useBox : unit.getUseBoxes()) {					
					if (useBox.getValue() instanceof InvokeExpr) {
						InvokeExpr invokeExpr = (InvokeExpr) useBox.getValue();
						if (invokeExpr.getMethodRef().name().matches("^findViewById$")) {
							Value widgetId = invokeExpr.getArg(0);
							usedWidgets.add(widgetId.toString());							
						}				
						System.out.println(invokeExpr.getMethodRef().name());
						System.out.println(invokeExpr.getMethodRef().getClass());
						System.out.println(invokeExpr.getMethodRef().getSubSignature());
						System.out.println(invokeExpr.getMethodRef().declaringClass());
						if (invokeExpr.getMethodRef().declaringClass().toString().equals("android.content.res.Resources") && invokeExpr.getMethodRef().name().equals("getString")) {
							Value widgetId = invokeExpr.getArg(0);
							usedWidgets.add(widgetId.toString());
						}
					}
				}
			}
			usedWidgets.remove(handlerWidgetId);
			handlerWidgetDirectAccessMap.put(methodName, usedWidgets);
			System.out.println(listener);
		} catch (Exception e) {
			System.err.println(listener);
		}
	}
	
	@Override
	public String getDependencyType() {
		return "Widget Handler Dependency";
	}

	public Map<String, HashSet<String>> getConditionsMap() {
		return null;
	}
}
