package edu.gmu.trimdroid.soot.utils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import soot.*;
import soot.jimple.InvokeExpr;
import soot.jimple.internal.JInstanceFieldRef;
import soot.tagkit.Tag;
import soot.util.Chain;

public class Util {

	public static boolean isInheritedFromClass(SootClass clazz, String superClassName) {
		SootClass superclass = clazz.getSuperclass();
		while (superclass != null && !superclass.getName().equals(Const.TYPE_OBJECT)) {
			if (superclass.getName().equals(superClassName))
				return true;
			superclass = superclass.getSuperclass();
		}
		return false;
	}

	public static void toXML(String outputDir, Object obj) throws JAXBException {
		JAXBContext jaxbContext = JAXBContext.newInstance(obj.getClass());
		Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
		jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
		jaxbMarshaller.marshal(obj, new File(outputDir));
	}

	public static SootClass getMainActivityClass(String appRoot) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		File manifestFile = getManifestFile(new File(appRoot));
		Document dom = db.parse(manifestFile);
		Element manifest = dom.getDocumentElement();
		String nameSpace = manifest.getAttribute("package");
		NodeList activities = manifest.getElementsByTagName("activity");
		for (int i = 0; i < activities.getLength(); i++) {
			Element activity = (Element) activities.item(i);
			NodeList actions = activity.getElementsByTagName("action");
			for (int j = 0; j < actions.getLength(); i++) {
				Element action = (Element) actions.item(j);
				if (action.getAttribute(Const.MANIFEST_NAME_ATTR).equals(Const.MANIFEST_MAIN_ACTION)) {
					String activityName = activity.getAttribute(Const.MANIFEST_NAME_ATTR);
					if (!activityName.startsWith("."))
						nameSpace += ".";
					String fullName = (activityName.startsWith(nameSpace) ? "" : nameSpace) + activityName;
					SootClass activityClass = Scene.v().loadClass(fullName, SootClass.BODIES);
					return activityClass;
				}
			}
		}
		return null;
	}

    public static String extractLayout(String fullName) throws ParserConfigurationException, IOException, SAXException {
        SootClass activityClass = Scene.v().loadClass(fullName, SootClass.BODIES);

        for(SootMethod sm:activityClass.getMethods()) {
            if (sm.getActiveBody() == null)
                continue;

            for(Unit unit:sm.getActiveBody().getUnits()) {
                if (unit == null)
                    continue;
                List<ValueBox> useBoxes = unit.getUseBoxes();
                for (ValueBox valueBox : useBoxes) {
                    Value value = valueBox.getValue();
                    if (value instanceof InvokeExpr) {
                        InvokeExpr invokeExpr = (InvokeExpr) value;
                        String methodName = invokeExpr.getMethod().getName();
                        if(methodName.equals("setContentView")) {
                            Value arg = invokeExpr.getArg(0);
                            if(arg.getType().toString().equals("int")) {
                                SootClass layoutRClass = findRClass("layout");
                                return integerId2StringId(arg.toString(), layoutRClass) + ".xml";
                            }

                        }
                        if(methodName.equals("addPreferencesFromResource")) {
                            Value arg = invokeExpr.getArg(0);
                            if(arg.getType().toString().equals("int")) {
                                SootClass layoutRClass = findRClass("layout");
                                String temp1 = integerId2StringId(arg.toString(), layoutRClass);
                                layoutRClass = findRClass("xml");                                
                                String temp2 = integerId2StringId(arg.toString(), layoutRClass);
                                String result = isNumeric(temp2) ? temp1:temp2;
                                return result + ".xml";
                            }

                        }
                    }
                }
            }
        }


        return "";
    }

    public static String extractAppPackage(String appRoot) throws ParserConfigurationException, IOException, SAXException {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        File manifestFile = getManifestFile(new File(appRoot));
        Document dom = db.parse(manifestFile);
        Element manifest = dom.getDocumentElement();
        String appPackage = manifest.getAttribute("package");

        return appPackage;
    }

	public static File getManifestFile(File appRoot) {
		File[] projectFiles = appRoot.listFiles();
		for (File projectFile : projectFiles) {
			if (Const.MANIFEST_FILE_NAME.equals(projectFile.getName()))
				return projectFile;
		}
		for (File projectFile : projectFiles) {
			if (projectFile.isDirectory()) {
				File result = getManifestFile(projectFile);
				if (result != null)
					return result;
			}
		}
		return null;
	}

	public static String getLibNames(String libFolder) {
		StringBuilder sb = new StringBuilder();
		File[] libs = new File(libFolder).listFiles();
		for (File lib : libs) {
			if (lib.getName().endsWith(".jar")) {
				if (sb.length() > 0)
					sb.append(":");
				sb.append(lib.getAbsolutePath());
			}
		}
		return sb.toString();
	}

	public static SootClass findRClass(String subclass) {
		for (SootClass clazz : Scene.v().getApplicationClasses()) {
			if (clazz.getShortName().equals("R$"+ subclass))
				return clazz;
		}
		return null;
	}

	public static String integerId2StringId(Object id, SootClass r) {
		if (id == null)
			return "NULL";
		if (r == null)
			return String.valueOf(id);
		Chain<SootField> fields = r.getFields();
		for (SootField sootField : fields) {
			List<Tag> tags = sootField.getTags();
			for (Tag tag : tags) {
				String tagValue = tag.toString();
				if (tagValue.startsWith("ConstantValue:")) {
					tagValue = tagValue.split(":")[1].trim();
					if (id.toString().equals(tagValue))
						return sootField.getName();
				}
			}
		}
		return String.valueOf(id);
	}

	public static Unit findClosestDefPointBackward(Unit usedPoint, SootFieldRef fieldRef, Chain<Unit> units) {
		Unit pred = null;
		while ((pred = units.getPredOf(usedPoint)) != null) {
			for (ValueBox defBox : pred.getDefBoxes()) {
				if (defBox.getValue() instanceof JInstanceFieldRef) {
					JInstanceFieldRef ref = ((JInstanceFieldRef) (defBox.getValue()));
					if (ref.getFieldRef().toString().equals(fieldRef.toString())) {
						return pred;
					}
				}
			}
			usedPoint = pred;
		}
		return null;
	}

	public static Unit findClosestDefPointBackward(Unit usedPoint, Value var, Chain<Unit> units) {
		Unit pred = null;
		while ((pred = units.getPredOf(usedPoint)) != null) {
			for (ValueBox defBox : pred.getDefBoxes()) {
				if (defBox.getValue().equals(var)) {
					return pred;
				}
			}
			usedPoint = pred;
		}
		return null;
	}

	public static Unit findClosestUsePointForward(Unit defPoint, SootFieldRef fieldRef, Chain<Unit> units) {
		Unit succ = null;
		while ((succ = units.getSuccOf(defPoint)) != null) {
			for (ValueBox useBox : succ.getUseBoxes()) {
				if (useBox.getValue() instanceof JInstanceFieldRef) {
					JInstanceFieldRef ref = ((JInstanceFieldRef) (useBox.getValue()));
					if (ref.getFieldRef().toString().equals(fieldRef.toString())) {
						return succ;
					}
				}
			}
			defPoint = succ;
		}
		return null;
	}

	public static Unit findClosestUsePointForward(Unit defPoint, Value var, Chain<Unit> units) {
		Unit succ = null;
		while ((succ = units.getSuccOf(defPoint)) != null) {
			for (ValueBox useBox : succ.getUseBoxes()) {
				if (useBox.getValue().equals(var)) {
					return succ;
				}
			}
			defPoint = succ;
		}
		return null;
	}

	public static Unit findClosestUsePointBackward(Unit defPoint, Value var, Chain<Unit> units) {
		Unit pred = null;
		while ((pred = units.getPredOf(defPoint)) != null) {
			for (ValueBox useBox : pred.getUseBoxes()) {
				if (useBox.getValue().equals(var)) {
					return pred;
				}
			}
			defPoint = pred;
		}
		return null;
	}

	public static InvokeExpr getInvokedMethod(Unit unit) {
		if (unit == null)
			return null;
		List<ValueBox> useBoxes = unit.getUseBoxes();
		for (ValueBox valueBox : useBoxes) {
			Value value = valueBox.getValue();
			if (value instanceof InvokeExpr) {
				return (InvokeExpr) value;
			}
		}
		return null;
	}

    public static void copyFile(String source, String dest) throws IOException {
        Files.copy(Paths.get(source), Paths.get(dest), StandardCopyOption.REPLACE_EXISTING);
    }
    
    public static boolean isNumeric(String s) {  
        return s.matches("[0-9]+");  
    }  

}
